/*	$LAAS: libdtm.h,v 1.8 2008/02/15 13:05:43 adegroot Exp $
 *
 * Copyright (c) 2000 LAAS/CNRS                       --  Fri Aug  4 2006
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */
#ifndef H_CALIFE_DTM
#define H_CALIFE_DTM

#include <stdbool.h>

/* includes from other libraries */
#include <t3d/t3d.h>
#include <libimages3d.h>

#ifdef __cplusplus
extern "C" {
#endif

/* Différentes énumérations pour les flags des cellules (aussi utilisées pour
   l'entête des images */
typedef enum {EMPTY, FILLED} cell_state;
typedef enum {FLAT, VERTICAL, UNKNOW} height_state;
typedef enum {STATIC, DYNAMIC} moving_state;

/* Différentes énumérations pour l'entête du DTM (pas utilisées pour les
   cellules) */
typedef enum {DTM_FIXED, DTM_SLIDING} dtm_type;
typedef enum {X_LIGNES, X_COLONNES} dtm_stockage;
typedef enum {NIL_CELLS, CELLS_ALLOUEES, CELLS_DETERMINEES} cells_infos;
typedef enum {NIL_DTM_BORNES, BORNES_DTM_CALCULEES} dtm_bornes_info;
typedef enum {NIL_LUMINANCES, LUMINANCES_DETERMINEES} dtm_luminances_info;
typedef enum {GEO_REFERENCE_SET, NO_GEO_REFERENCE} dtm_geo_reference;
typedef enum {NO_GEO_ORIENTATION, GEO_NORTH, GEO_WEST, 
	      GEO_SOUTH, GEO_EAST} dtm_geo_orientation;
typedef enum {DTM_AVERAGE, DTM_WEIGTHED_AVERAGE, DTM_MAX} DTM_FUSION;

/* Entêtes des fichiers de sauvegarde de la structure 'DTM' */
#define DTM_ASCII_FF		"DTM_ASCII_file_format"
#define DTM_P3D_ASCII_FF	"DTM_ASCII_P3D_file_format"
#define DTM_BINARY_FF		"DTM_BINARY_file_format"

typedef struct dtm_cell_info {
  int		nb_points;	/* Le nombre de pts 3D inclus */
  short		luminance;	/* La luminance moyenne */
  short		firstDate;	/* Première date (num. d'im3d) de mise à jour */ 
  short		lastDate;	/* Dernière date (num. d'im3d) de mise à jour */ 
  float		z_moyen;	/* L'élévation moyenne */
  float		z_max;		/* Elévation max perçue */
  float		z_min;		/* Elévation min perçue */
  float		sigma_z;	/* Écart type sur l'élévation */
  float		confidence;	/* Confiance sur les infos z et luminance */
  float		weight;		/* Poids cumulé des points fusionnés */
} DTM_CELL_INFO;
/***********************************************************************
 *
 * La structure 'dtm_cell' : "pixels" du DTM
 *
 ***********************************************************************/

typedef struct dtm_cell
{
  /* Quelques champs dans lesquel on peut mettre tout un tas
     d'informations. Voir dans la séquence des énumérations ci-dessus la
     signification des valeurs correspondantes */
  unsigned       state:1;	/* L'état du point */

  height_state	hstate;		/* Surface horizontale ou verticale */
  moving_state  mstate;		/* Objet dynamique ou nom */
  int may_be_visible;		/* la zone est-elle potentiellement visible (et par
							   combien de camera*/

  DTM_CELL_INFO current_info; /* info courante */
  DTM_CELL_INFO floor_info;	  /* info sur le sol si connu */

}  DTM_CELL;

#define		NOT_SEEN	 0x0000
#define		SEEN_LEFT	 0x0001
#define		SEEN_RIGHT	 0x0002
#define		SEEN_BOTH	 (SEEN_LEFT|SEEN_RIGHT)

#ifdef APPLY_VISIBILITY
static inline bool 
dtm_visible_cell(DTM_CELL * cell)
{
	return (cell->may_be_visible == SEEN_BOTH);
};
#else
static inline bool
dtm_visible_cell(DTM_CELL* cell)
{
	/*
	 * dtm_visible_cell is called to check if we must have some data here
	 * returning always false let the algorithm run as expected, if they are no
	 * new data, don't do anything for this point
	 */
	return false;
};
#endif

typedef enum dtm_strategy {
	FIXED_TRESHOLD,
	MEAN_TRESHOLD,
} dtm_strategy;

typedef struct dtm_dynamic_fuse_strategy {
	dtm_strategy strat;
	double treshold;
} dtm_dynamic_fuse_strategy;

/***********************************************************************
 *
 * Petite structure qui contient les différents champs d'une position
 * géodesique. Utilisée pour référencer l'origine d'un mnt.
 *
 ***********************************************************************/

typedef struct dtm_geodesic_point
{
  /* On met un peu tous les champs habituels à une coordonnée
     géodésique (en WGS84) */ 
  double        lat, lon, alt;	/* LLA coordinates */

  int           gridZoneDesignation;
  int           gridZone;
  double        easting, northing; /* UTM coordinates */
  double        nextEastMeridian; /* This one is necessary for UTM <->
				     LLA conversions. */

} DTM_GEODESIC_POINT;

/***********************************************************************
 *
 * La structure 'dtm'. Les différentes conventions sont identiques à celles
 * que matthieu à définies pour gdhe (in 'gdhe_terrain.h')
 *
 ***********************************************************************/

typedef struct dtm
{
  /* On commence par tout un tas de caractéristiques générales du modèle de
     terrain */ 
  dtm_type              type;	/* Indicates whether the DTM is moving along
				   with the rover or not */
  dtm_stockage          stockage;
  cells_infos           cells;
  dtm_bornes_info       bornes;
  dtm_luminances_info   luminance;
  dtm_geo_reference     geoReference;
  dtm_geo_orientation   geoOrientation;

  int		nb_im3d;	/* Nombres d'images 3D qu'on à mis dans le DTM */

  /* Suivent tout un tas d'infos sur la géométrie du DTM */
  DTM_GEODESIC_POINT geodesicOrigin; /* Coordonnées géodésiques du
					point (0,0) du DTM */
  double            x_origine;
  double            y_origine;	/* Coordinates of the top-left corner point of
				   the first DTM cell (the one with (0,0)
				   integer coordinates) in the global frame */ 

  int               i_origine;
  int               j_origine;	/* Coordinates in pixels of the cell that
				   contains the origin of the global frame
				   (they can be negative or larger than the DTM
				   size) */
  
  double            scale_x;
  double            scale_y;	/* Taux de discrétisation du DTM (1 pixel
				   équivaut à 'scale' mètre */

  double            size_x;
  double            size_y;	/* Taille du DTM (en mètres oeuf corse) */

  int               nblig;
  int               nbcol;	/* Nombre de lignes et de colonnes du DTM */
  
  int               nb_cells;	/* Nombre de cellules pleines */

  /* All the fields that follow are bounds on the cells' attributes */
  int               i_min, i_max;
  int               j_min, j_max; /* Indices min et max qui englobent l'ensemble
				     des cellules perçues */

  double            z_min, z_max; /* Bornes de l'élévation */
  int		    nbPtsMin, nbPtsMax;
  double            sigmaMin, sigmaMax;
  short             colorMin, colorMax;

  /* Enfin les données */
  DTM_CELL*         cells_tab;	/* le tableau des cellules : il est rangé
				   comme les ima1 (ie. comme un tableau en C)
				   ... */ 

   /* Internal data */
   void *	    data;	/* pointer to private quad structure for
				   interpolation */
}  DTM;

/* macro d'acces à une cellule du DTM */
#define dtm_one_cell(a,i,j) (((a)->cells_tab) + ((i) * ((a)->nbcol)) + (j))

/*
 * Structures for polygons filling by interpolation ---------------------
 */

/* A span is a line of pixels in the dtm. It spans from `min' to `max'
 * indexes */
typedef struct DtmQuadSpan {
   char count;			/* number of pixels in this span */
   int min, max;		/* min and max indexes */
   float zmin, zmax;		/* elevation at min and max index */
   float imin, imax;		/* luminance at min and max index */
} DtmQuadSpan;

/* A quad is a polygon with at most 4 edges. Vertices are numbered from 0
 * to 3. This structure contains arrays too, which are used by the
 * interpolation routines (dtmFuseQuad) to buffer vertices coordinates
 * and other time consuming computations.  */

typedef struct DtmQuad {
   char d[4];			/* valid vertices (boolean) */
   int x[4], y[4];		/* vertices coordinates */
   float z[4], i[4];		/* pixel heigths and luminances */

   int spans;			/* number of spans */
   DtmQuadSpan *span;		/* array of spans */

   int vertices;		/* number of vertices in the array below
				 * (normally the dtm line length) */
   char *vd;			/* buffered `d' array */
   int *vx, *vy;		/* buffered `x' and `y' arrays */
   float *vz, *vi;		/* buffered `z' and `i' arrays */
} DtmQuad;


/* Library error codes */
int dtmPrintError(int num, char *errorCode, char *fileName, int lineNumber);
#define dtmError(num) (dtmPrintError(num, #num, __FILE__, __LINE__))
#define dtm_Id  532
#define DTM_OK 0
#define DTM_ErrNum(mod,num) ((((mod) & 0xFFFF) << 16) | ((num) & 0xFFFF))


/* Les pbs d'allocation */
#define ERR_DTM_MALLOC				DTM_ErrNum(dtm_Id, 0)
#define ERR_DTM_MALLOC_CELLS			DTM_ErrNum(dtm_Id, 1)
#define ERR_DTM_CELLS_DEJA_ALLOUEES		DTM_ErrNum(dtm_Id, 2)
#define ERR_DTM_CELLS_NON_ALLOUEES		DTM_ErrNum(dtm_Id, 3)
#define	ERR_DTM_ALLOC_POSTER			DTM_ErrNum(dtm_Id, 4)

/* Les paramètres invalides */
#define ERR_DTM_SIZE				DTM_ErrNum(dtm_Id, 10)
#define ERR_DTM_SCALE				DTM_ErrNum(dtm_Id, 11)
#define ERR_DTM_TYPE_STOCKAGE			DTM_ErrNum(dtm_Id, 12)
#define ERR_DTM_INDICES				DTM_ErrNum(dtm_Id, 13)
#define ERR_DTM_IM3D_NON_CALCULEE		DTM_ErrNum(dtm_Id, 14)
#define ERR_DTM_IM3D_COORDONNEES		DTM_ErrNum(dtm_Id, 15)
#define ERR_DTM_TAILLES_IMAGES			DTM_ErrNum(dtm_Id, 16)
#define ERR_DTM_VIDE				DTM_ErrNum(dtm_Id, 17)
#define ERR_DTM_NO_LUMINANCE			DTM_ErrNum(dtm_Id, 18)
#define ERR_DTM_TAUX_REDUCTION			DTM_ErrNum(dtm_Id, 19)
#define ERR_DTM_ODD_SIZE			DTM_ErrNum(dtm_Id, 20)
#define ERR_DTM_NUM_POSITION			DTM_ErrNum(dtm_Id, 21)
#define ERR_DTM_OUT_OF_DTM			DTM_ErrNum(dtm_Id, 22)

/* Erreurs relatives à l'affichage */
#define ERR_DTM_CREATION_IMA1			DTM_ErrNum(dtm_Id, 30)
#define ERR_DTM_DECODAGE_IMA1			DTM_ErrNum(dtm_Id, 31)
#define ERR_DTM_CREATION_IMA2			DTM_ErrNum(dtm_Id, 32)
#define ERR_DTM_DECODAGE_IMA2			DTM_ErrNum(dtm_Id, 33)
#define ERR_DTM_CREATION_IMA4			DTM_ErrNum(dtm_Id, 34)
#define ERR_DTM_DECODAGE_IMA4			DTM_ErrNum(dtm_Id, 35)
#define ERR_DTM_DECODAGE_IMA8			DTM_ErrNum(dtm_Id, 36)
#define ERR_DTM_DECODAGE_R8			DTM_ErrNum(dtm_Id, 37)
#define ERR_DTM_NOT_OKFEN			DTM_ErrNum(dtm_Id, 38)

/* Erreurs de dépassement */
#define ERR_DTM_OVER_NB_POSITIONS_MAX		DTM_ErrNum(dtm_Id, 40)

/* Erreurs diverses */
#define ERR_DTM_ROBOT_OUT_OF_DTM		DTM_ErrNum(dtm_Id, 50)
#define ERR_DTM_BAD_PARAMS			DTM_ErrNum(dtm_Id, 51)
#define ERR_DTM_NO_CELLS_UPDATED		DTM_ErrNum(dtm_Id, 52)

/* Erreurs relatives à des fichiers */
#define ERR_DTM_OPEN_FILE			DTM_ErrNum(dtm_Id, 60)
#define ERR_DTM_READ_FILE			DTM_ErrNum(dtm_Id, 61)
#define ERR_DTM_FILEFORMAT			DTM_ErrNum(dtm_Id, 62)
#define ERR_DTM_WRITE_FILE			DTM_ErrNum(dtm_Id, 63)

/* Incompatible dtm to merge */
#define ERR_DTMS_INCOMPATIBLES			DTM_ErrNum(dtm_Id, 70)

/* Erreurs diverses */
#define ERR_DTM_CAS_IMPOSSIBLE			DTM_ErrNum(dtm_Id, 98)
#define ERR_DTM_FCT_NON_IMPLEMENTEE		DTM_ErrNum(dtm_Id, 99)


/* Library function prototypes */
DTM* copy_dtm ( DTM* mnt );
DTM* create_dtm ( void );
void clear_dtm ( DTM* mnt );
int dtmAllocQuads ( DtmQuad *quad, int width );
void dtmDeleteQuads ( DtmQuad *quad );
int dtmInitQuads ( DtmQuad *quad );
int dtmMakeQuad ( DtmQuad *quad, int *ymin, int *ny );
DTM* prepare_dynamic_fuse(T3D* sensorLeftToMain, T3D* sensorRightToMain, double *intrinsincParamsRectified, int nbLines, int nbColumns, int reduction, double max_depth, double scale_x, double scale_y);
void clean_dynamic_fuse(DTM * dtm);
dtm_dynamic_fuse_strategy dtm_dynamic_get_strategy(DTM* dtm);
void dtm_dynamic_set_strategy(DTM* dtm, dtm_dynamic_fuse_strategy);

T3D * dtmPoseRefinement ( DATA_IM3D *im3d, DTM *mnt, int step );
DTM* dtm_autocrop ( DTM* mnt );
int dtm_boundFirstDate ( DTM* mnt, int maxFirstDate );
int dtm_boundLastDate ( DTM* mnt, int maxLastDate );
int dtm_checkInitParams ( double scale_x, double scale_y, int nb_x, int nb_y, dtm_stockage stockage );
DTM* dtm_crop ( DTM* mnt, int iMin, int iMax, int jMin, int jMax );
int dtm_fuseData ( DATA_IM3D* im3d, T3D* RobotRed_2_Start, DTM* mnt, DTM_FUSION fusion, int color);
int dtm_dynamic_fuse ( DATA_IM3D* im3d, T3D* RobotRed_2_Start, T3D* sensorToOrigin, DTM* mnt, DTM* tmp_mnt, int color);
int dtm_fuseQuadIm3dIma1 ( DATA_IM3D* im3d, T3D* RobotRed_2_Start, DTM* mnt, int color, unsigned char* q_image, int nblig, int nbcol );
int dtm_getElevation ( DTM* mnt, double x, double y, double* z );
int dtm_getUpdateSize ( DTM* mnt, int lastDate, int* nbUpdatedCells );
void dtm_infoGeodesicPos ( DTM_GEODESIC_POINT* pos );
void dtm_intersectLineDtm ( DTM* dtm, int i1, int j1, int i2, int j2, double betaX, double translX, double betaZ, double translZ, int* intersection, int* iInter, int* jInter );
int dtm_minmax ( DTM* mnt );
DTM* dtm_readArcGisAscii ( FILE* file );
int dtm_writeArgGisAscii (DTM* dtm, FILE* file, int lambert);
DTM* dtm_readAsciiDtm ( FILE* file );
DTM* dtm_readAsciiUsgs ( FILE* file, int xyStep, double scale, double zScale );
int dtm_setLLAGeodesicReference ( DTM* dtm, int consideredPoint, double lat, double lon, double alt );
int dtm_setUTMGeodesicReference ( DTM* dtm, int consideredPoint, double easting, double northing, double alt, double nextEast );
DATA_IM3D* dtm_simulStereo ( DTM* dtm, T3D* cam2world, double* intrinsincParams, int nbLinesInit, int nbColumnsInit, int reduction, double maxDistance);
void dtm_wgs84lla2utm ( double lon_rad, double lat_rad, double lon0_rad, double* easting, double* northing );
void dtm_wgs84utm2lla ( double x, double y, double meridian, double* lat, double* lon );
void dtm_lla2LambertIIe (double latitudeWGS84, double longitudeWGS84, double* latitudeLambert, double* longitudeLambert);
int dtm_writeAsciiDtm ( DTM* mnt, FILE* file );
int destroy_dtm ( DTM* mnt );
DTM* down_sample_dtm ( DTM* mnt, int reduction, int moyennage );
DTM* ima8_to_dtm ( double* q_image, int nblig, int nbcol, double scale_x, double scale_y, int stockage );
int info_dtm ( DTM* mnt );
int info_dtm_cell ( DTM* mnt, int i, int j );
int init_dtm_cells ( DTM* mnt, double scale_x, double scale_y, int nb_x, int nb_y, double x_origine, double y_origine, dtm_type type, dtm_stockage stockage );
int print_dtm_bornes ( DTM* mnt );
DTM* dtm_readBinaryDtm( FILE* file );
int void_dtm_cell ( DTM* mnt, int i, int j );
int dtm_writeBinaryDtm( DTM* mnt, FILE* file );
int dtm_slideDtm (DTM* dtm,T3D* RobotRed_2_Start);
int dtm_jumpDtm (DTM* dtm, T3D* posOld2Origin, T3D* posNew2Origin);

// Some useful macro
#ifndef MAX
#define MAX(a,b) ((a) > (b) ? (a) : (b))
#endif

#ifndef MIN
#define MIN(a,b) ((a) < (b) ? (a) : (b))
#endif

#ifndef CONVRDG
#define CONVRDG 180./M_PI
#endif

#ifndef CONVDGR
#define CONVDGR M_PI/180.
#endif

#ifndef DEG_TO_RAD
#define DEG_TO_RAD(x) ((x)*CONVDGR)
#endif

#ifndef RAD_TO_DEG
#define RAD_TO_DEG(x) ((x)*CONVRDG)
#endif

#ifndef M_TO_MM
#define M_TO_MM(a) ((a) * 1000.0)
#endif

#ifndef MM_TO_M
#define MM_TO_M(a) ((a) * 0.001)
#endif

#ifdef __cplusplus
}
#endif

#endif /* H_CALIFE_DTM */
