/* Les includes génériques de C */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* structures, protos and errors of the lib */
#include "libdtm.h"

#define DTM_ARCGIS_NO_DATA_VALUE  (-9999)

/**********************************************************************
 *
 * Read a DTM in ASCII ArcGis format
 *
 **********************************************************************/

DTM*
dtm_readArcGisAscii (FILE*	file)

{
  int		i, j;		/* Pour boucler */
  DTM*		mnt;
  DTM_CELL*	cell;

  /* Variables read in the file */
  int nbCols, nbRows;
  float xLL, yLL;
  float cellSize;
  int noDataValue;
  float altitude;

  /* Creates the DTM */
  mnt = create_dtm ();
  if (mnt == NULL) {
    dtmError (ERR_DTM_MALLOC);
    return (NULL);
  }

  /* Read the file header */
  if (fscanf (file, "ncols         %d\n", &nbCols) != 1) {
    dtmError (ERR_DTM_FILEFORMAT);
    return (NULL);
  }
  if (fscanf (file, "nrows         %d\n", &nbRows) != 1) {
    dtmError (ERR_DTM_FILEFORMAT);
    return (NULL);
  }
  if (fscanf (file, "xllcorner     %f\n", &xLL) != 1) {
    dtmError (ERR_DTM_FILEFORMAT);
    return (NULL);
  }
  if (fscanf (file, "yllcorner     %f\n", &yLL) != 1) {
    dtmError (ERR_DTM_FILEFORMAT);
    return (NULL);
  }
  if (fscanf (file, "cellsize      %f\n", &cellSize) != 1) {
    dtmError (ERR_DTM_FILEFORMAT);
    return (NULL);
  }
  if (fscanf (file, "NODATA_value  %d\n", &noDataValue) != 1) {
    dtmError (ERR_DTM_FILEFORMAT);
    return (NULL);
  }

  /* Initializes the DTM according the file header */
  mnt->cells = NIL_CELLS;	/* In order to not get an error from the
				   init_dtm_cells fct  */
  if (init_dtm_cells (mnt, (double) cellSize, (double) cellSize,
		      nbRows, nbCols,  0, 0,
		      /* mnt->x_origine, mnt->y_origine, */
		       DTM_FIXED, X_LIGNES) != DTM_OK) {
    dtmError (ERR_DTM_MALLOC);
    return (NULL);
  }

  /* Note: throughout the library, we consider that the geodesic reference
     correspond to the corner of the LowerLeft cell - not its center */

  /* Set the geo-reference correctly */
  mnt->geodesicOrigin.easting = xLL;
  mnt->geodesicOrigin.northing = yLL;
  dtm_wgs84utm2lla (mnt->geodesicOrigin.easting, 
		    mnt->geodesicOrigin.northing,
		    mnt->geodesicOrigin.nextEastMeridian,
		    &(mnt->geodesicOrigin.lat),
		    &(mnt->geodesicOrigin.lon));
  mnt->geoReference = GEO_REFERENCE_SET;
  mnt->geoOrientation = GEO_EAST;
 
  /* Now scans the columns and lines */
  cell = mnt->cells_tab;
  for (i = 0; i < nbRows; i++) {
    for (j = 0; j < nbCols; j++) {
      if (fscanf (file, "%f ", &altitude) != 1) {
	dtmError (ERR_DTM_FILEFORMAT);
	return (NULL);
      }
      if (altitude != noDataValue) {
	cell->state = FILLED;
	cell->hstate = FLAT;
	cell->mstate = STATIC;
	cell->current_info.nb_points = 1;
	cell->current_info.luminance = 0;
	cell->current_info.firstDate = 1;
	cell->current_info.lastDate = 1;
	cell->current_info.sigma_z = 0.0;
	cell->current_info.z_moyen = altitude;
      }
      else
	cell->state = EMPTY;

      cell++;
    }
    fscanf (file, "\n");
  }    

  mnt->geodesicOrigin.alt = mnt->cells_tab->current_info.z_moyen;

  /* Update some flags */
  mnt->cells = CELLS_DETERMINEES;

  return (mnt);
}

/**********************************************************************
 *
 * Writes a dtm file writen as .asc files
 *
 **********************************************************************/

int
dtm_writeArgGisAscii (DTM*	dtm,
		      FILE*	file,
		      int	lambert)
{
  int		i, j;		/* To loop on the DTM cells */
  double	xll, yll;	/* Coordinates of the center of the Lower Left
				   cell (that is, the southWest in the
				   geo-referenced frame) written in the file
				   header */
  DTM_CELL*	cell;

  /* Write the header */
  fprintf (file, "ncols         %d\n", dtm->nbcol);
  fprintf (file, "nrows         %d\n", dtm->nblig); 
  /* No geo-reference is set, set it to (0,0) and print a warning */
  if (dtm->geoReference != GEO_REFERENCE_SET) {
    printf ("### Warning: no geo-reference is known for this DTM -- setting it to (0,0)\n");
    xll = 0.0;
    yll = 0.0;
  } else {
    /* Given that x=East and y=North, the Lower Left cell in the geo-referenced
       frame is the top-left cell in the dtm image frame -- that is, the cell of
       coordinates (0,0) */
    if (lambert == 0) {
      /* Write the utm coordinates */
      xll = dtm->geodesicOrigin.easting;
      yll = dtm->geodesicOrigin.northing;
    } else { /* Compute xll and yll as lambert coordinates */
      dtm_lla2LambertIIe (dtm->geodesicOrigin.lat, 
			  dtm->geodesicOrigin.lon,
			  &xll, &yll);
    }
  }
  fprintf (file, "xllcorner     %.2f\n", xll);
  fprintf (file, "yllcorner     %.2f\n", yll);
  fprintf (file, "cellsize      %.2f\n", dtm->scale_x);
  fprintf (file, "NODATA_value  %d\n", DTM_ARCGIS_NO_DATA_VALUE);

  /* Write all the points altitude */
  cell = dtm->cells_tab;
  for (i = 0; i < dtm->nblig; i++) {
    for (j = 0; j < dtm->nbcol; j++) {
      if (cell->state != EMPTY)
	fprintf (file, "%f ", cell->current_info.z_moyen);
      else
	fprintf (file, "%d ", DTM_ARCGIS_NO_DATA_VALUE);
      cell++;
    }
    fprintf (file, "\n");
  }

  return (DTM_OK);
}
