/* Les includes génériques de C */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <float.h>
#include <limits.h>
#include <math.h>

/* Library includes */
#include "libdtm.h"

/**
 *
 * Simple allocation d'une structure 'MNT' (le tableau de points n'est
 * pas alloué)
 *
 */

DTM*	
create_dtm ()

{
  DTM*	mnt;
  
  mnt = (DTM*) malloc (sizeof (DTM));

  if (mnt == NULL) {
    dtmError (ERR_DTM_MALLOC);
    return (NULL);
  }

  /* On positionne les flags comme il faut */
  mnt->cells = NIL_CELLS;
  mnt->bornes = NIL_DTM_BORNES;
  mnt->stockage = X_LIGNES;
  mnt->luminance = NIL_LUMINANCES;
  mnt->geoReference = NO_GEO_REFERENCE;
  mnt->geoOrientation = NO_GEO_ORIENTATION;

  /* Quelques initialisations (un memset serait équivalent) */
  mnt->geodesicOrigin.lat = 0.0;
  mnt->geodesicOrigin.lon = 0.0;
  mnt->geodesicOrigin.alt = 0.0;
  mnt->geodesicOrigin.gridZoneDesignation = 0;
  mnt->geodesicOrigin.gridZone = 0;
  mnt->geodesicOrigin.easting = 0.0;
  mnt->geodesicOrigin.northing = 0.0;
  mnt->geodesicOrigin.nextEastMeridian = 0.0;

  mnt->x_origine = 0;
  mnt->y_origine = 0;

  mnt->scale_x = 0;
  mnt->scale_y = 0;

  mnt->size_x = 0;
  mnt->size_y = 0;

  mnt->nblig = 0;
  mnt->nbcol = 0;

  mnt->nb_cells = 0;

  mnt->nb_im3d = 0;

  mnt->i_max = - INT_MAX;
  mnt->i_min = INT_MAX;
  mnt->j_max = - INT_MAX;
  mnt->j_min = INT_MAX;

  mnt->z_max = - FLT_MAX;
  mnt->z_min = FLT_MAX;

  mnt->nbPtsMax = - INT_MAX;
  mnt->nbPtsMin = INT_MAX;

  mnt->sigmaMax = - FLT_MAX;
  mnt->sigmaMin = FLT_MAX;

  mnt->colorMin = SHRT_MAX;
  mnt->colorMax = - SHRT_MAX;

  mnt->cells_tab = NULL;

  mnt->data = NULL;

  return (mnt);
}

/***********************************************************************
 *
 * Verification de la validité des paramètres d'initialisation
 *
 ***********************************************************************/

int
dtm_checkInitParams ( double		scale_x,
			double      	scale_y,
			int		nb_x,
			int		nb_y,
			dtm_stockage stockage)

{
  if ((scale_x <= 0.0) || (scale_y <= 0.0))
    return dtmError(ERR_DTM_SCALE);
  
  if ((nb_x <= 0) || (nb_y <= 0))
    return dtmError(ERR_DTM_SIZE);
  
  if (stockage != X_LIGNES)
    return dtmError(ERR_DTM_TYPE_STOCKAGE);
  
  return (DTM_OK);
}

/**
 *
 * Allocates the cell array of a 'DTM' structure
 *
 * @param DTM The exsiting DTM structure
 * @param scale_x
 */

int
init_dtm_cells (DTM*		mnt,
		double		scale_x,
		double      	scale_y,
		int		nb_x,
		int		nb_y,
		double		x_origine,
		double      	y_origine,
		dtm_type		type,
		dtm_stockage	stockage)
     
{
  /* Validité des paramètres */
    if (dtm_checkInitParams (scale_x, scale_y, 
			     nb_x, nb_y, stockage) != DTM_OK)
    return (-1);

  /* On vérifie que le tableau des cellules n'est pas déjà alloué */
  if (mnt->cells != NIL_CELLS)
    return dtmError(ERR_DTM_CELLS_DEJA_ALLOUEES);
  
  /* Détermination de la taille du MNT */
  mnt->size_x = nb_x * scale_x;
  mnt->size_y = nb_y * scale_y;

  /* Mise à jour de l'entête (paramètres et des flags) */
  mnt->scale_x = scale_x;
  mnt->scale_y = scale_y;

  mnt->x_origine = x_origine;
  mnt->y_origine = y_origine;

  mnt->cells = CELLS_ALLOUEES;
  mnt->bornes = NIL_DTM_BORNES;
  mnt->type = type;
  mnt->stockage = stockage;

  switch (stockage) {
  case X_LIGNES:
    mnt->nblig = nb_x;
    mnt->nbcol = nb_y;
    break;
  case X_COLONNES:
    mnt->nbcol = nb_x;
    mnt->nblig = nb_y;
    break;
  default:
    return dtmError(ERR_DTM_CAS_IMPOSSIBLE);
/*     break; */
/* Unreached */
  }

  /* Allocation du tableau de cellules */
  mnt->cells_tab = (DTM_CELL*) malloc (mnt->nblig * mnt->nbcol * sizeof (DTM_CELL));
  
  if (mnt->cells_tab == NULL)
    return dtmError(ERR_DTM_MALLOC_CELLS);
  

  /* Détermination des coordonnées pixels de l'origine du repère global */
  switch (mnt->stockage) {
  case X_LIGNES:
    mnt->i_origine = (int) (floor (- x_origine / scale_x));
    mnt->j_origine = (int) (floor (- y_origine / scale_y));
    break;
  case X_COLONNES:
    mnt->j_origine = (int) (floor (- x_origine / scale_x));
    mnt->i_origine = (int) (floor (- y_origine / scale_y));
    break;
  default:
    return dtmError(ERR_DTM_CAS_IMPOSSIBLE);
/*     break; */
/* Unreached */
  }
  
  /* On met tout le tableau des cellules à zéro : ça a l'avantage
     d'initialiser les flags des cellules comme il faut */
  memset ( (DTM_CELL*) mnt->cells_tab, 0, mnt->nblig * mnt->nbcol * sizeof (DTM_CELL) );
  


  /* Install quad structure */
  mnt->data = malloc(sizeof(DtmQuad));
  if (!mnt->data) return dtmError(ERR_DTM_MALLOC_CELLS);
  
  dtmInitQuads(mnt->data);
  if (dtmAllocQuads(mnt->data, mnt->nbcol))
    return dtmError(ERR_DTM_MALLOC_CELLS);
  
  return (DTM_OK);
}


/**********************************************************************
 *
 * Destruction d'une structure 'DTM'
 *
 **********************************************************************/

int
destroy_dtm (DTM*	mnt)

{
  /* On commence par effacer les tableaux de données */
  if (mnt->cells != NIL_CELLS)
    free (mnt->cells_tab);

  /* Release quad storage */
  if (mnt->data) {
     dtmDeleteQuads(mnt->data);
     free(mnt->data);
     mnt->data = NULL;
  }

  /* On efface la structure elle-même */
  free (mnt);
		    
  return (DTM_OK);
}

/**********************************************************************
 *
 * Copie d'une structure 'DTM'
 *
 **********************************************************************/

DTM*
copy_dtm (DTM*	mnt)

{
  (void) mnt;
  printf ("ERR_DTM_FCT_NON_IMPLEMENTEE\n");
  return (NULL);
}

/**********************************************************************
 * Clear the content of dtm, aka remove information stored in it
 * Only the content of cells is killed, information about size and other things
 * are keeped
 */
void
clear_dtm( DTM* dtm)
{
  dtm->nb_im3d = 0;

  dtm->nb_cells = 0;

  dtm->i_max = - INT_MAX;
  dtm->i_min = INT_MAX;
  dtm->j_max = - INT_MAX;
  dtm->j_min = INT_MAX;

  dtm->z_max = - FLT_MAX;
  dtm->z_min = FLT_MAX;

  dtm->nbPtsMax = - INT_MAX;
  dtm->nbPtsMin = INT_MAX;

  dtm->sigmaMax = - FLT_MAX;
  dtm->sigmaMin = FLT_MAX;

  dtm->colorMin = SHRT_MAX;
  dtm->colorMax = - SHRT_MAX;

  /* On met tout le tableau des cellules à zéro : ça a l'avantage
     d'initialiser les flags des cellules comme il faut */
  memset ( (DTM_CELL*) dtm->cells_tab, 0, dtm->nblig * dtm->nbcol * sizeof (DTM_CELL) );

  /* XXX What to do with internal data ? */
}

