/* Les includes génériques de C */
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>

/* Les structures, protos et erreurs de la lib */
#include "libdtm.h"
 
/* Important : voici un point sur les relations entre les coordonnées entières
   (indices des pixels du DTM) et doubles (ou réelles quoi ;-). 

   - un pixel (i, j) correspond au carré centré en x = ((i + 0.5) * scale_x) +
   mnt->x_origine, et y = ((j + 0.5) * scale_y) + mnt->y_origine.

   - Donc pour fusionner un point 3D (x, y), on trouve ses coordonnées (i, j)
   en faisant : i = (x - x_origine) / scale_x, puis en castant en int (idtm
   pour j). Ainsi c'est bien les points qui se projettent sur la surface du
   carré décrit par un pixel qui sont fusionnés dans ce pixel.

   - On appelle "origine du mnt" le coin en haut à gauche du pixel (0, 0).

   - C'est le coin en haut à gauche du pixel (0, 0) qui correspond aux champs
   'x_origine' et 'y_origine'.

   - Le pixel (i_origine, j_origine) est celui dont le coin haut-gauche a pour 
   coordonnées (0.0, 0.0) dans le repère global.

   */

static inline void
dtm_fusePoint (DTM_CELL*	cell,
	       DATA_PT3D*	pt3d,
	       int		date,
	       float		weight,
	       DTM_FUSION	fusionType,
	       int		color,
	       unsigned char	pix)

{
  float			old_z_moyen; 
  int			new_nb_points;
  float			new_z;

  cell->mstate = STATIC;
  if (cell->state == EMPTY) {
    cell->state = FILLED;
    cell->current_info.nb_points = 1;
    cell->current_info.firstDate = date;
    cell->current_info.lastDate = date;
    cell->current_info.z_moyen = pt3d->coord_3;
    cell->current_info.z_max = pt3d->coord_3;
    cell->current_info.z_min = pt3d->coord_3;
    cell->current_info.weight = weight;
    if (color != 0)
      cell->current_info.luminance = pix;
  }
  else { 
    cell->current_info.lastDate = date;
    cell->current_info.nb_points++;
    switch (fusionType) {
    case DTM_AVERAGE:
      old_z_moyen = cell->current_info.z_moyen;
      new_nb_points = cell->current_info.nb_points + 1;
      new_z = pt3d->coord_3;
      /* Update mean altitude */
      cell->current_info.z_moyen 
	= ((old_z_moyen * cell->current_info.nb_points) + new_z) 
	/ new_nb_points;
      /* Update sigma altitude */
      cell->current_info.sigma_z 
	= ((cell->current_info.nb_points 
	    * (cell->current_info.sigma_z + (old_z_moyen * old_z_moyen))
	    + (new_z * new_z)) / new_nb_points)
	- (cell->current_info.z_moyen * cell->current_info.z_moyen);
      
      cell->current_info.z_max = MAX(cell->current_info.z_max, new_z);
      cell->current_info.z_min = MAX(cell->current_info.z_min, new_z);
      
      /* Update mean luminance */
      if (color != 0)
	cell->current_info.luminance = 
	  (unsigned char) (((cell->current_info.luminance * cell->current_info.nb_points) + pix) 
			   / new_nb_points);
      
      break;
    case DTM_WEIGTHED_AVERAGE:
      ;
      break;
    case DTM_MAX:
      if (pt3d->coord_3 > cell->current_info.z_moyen) {
	cell->current_info.z_moyen = pt3d->coord_3;
	cell->current_info.lastDate = date;
	if (color != 0)
	  cell->current_info.luminance = pix;
      }
      break;
    default:
      break;
    }
  }
  
  return;
}

/**********************************************************************
 *
 * Simple remplissage d'un MNT par méthode moyenageuse, avec
 * éventuellement infos de couleur. 
 *
 **********************************************************************/

int
dtm_fuseData (DATA_IM3D*		im3d,
	      T3D*			RobotRed_2_Start,
	      DTM*			mnt,
	      DTM_FUSION		fusionType,
	      int			color)

{
  int			nbligIm3d;
  int			nbcolIm3d;
  int			i, j;	/* Pour boucler sur les points de l'im3d */
  DATA_PT3D*		pt3d;	/* Un point de l'im3d */
  short			date;	/* La date (num. d'im3d) de mise à jour lors
				   de l'appel de cette fct */

  int			i_mnt;
  int			j_mnt;	/* Indices d'une cellule dans le MNT */
  DTM_CELL*		cell;
  int			nblig_mnt;
  int			nbcol_mnt;
  unsigned char*	luminance;
  int			nb_pts3d_out; /* Les nombre de pts3d hors du MNT */

  /* valididité des paramtres */
  if (mnt->cells == NIL_CELLS)
    return dtmError (ERR_DTM_CELLS_NON_ALLOUEES);

  if (im3d->header.points != POINTS_CALCULES)
    return dtmError (ERR_DTM_IM3D_NON_CALCULEE);

  if (im3d->header.coordonnees != COORDONNEES_CARTESIENNES)
    return dtmError (ERR_DTM_IM3D_COORDONNEES);

  if (mnt->stockage != X_LIGNES) 
    return dtmError (ERR_DTM_TYPE_STOCKAGE);

  /* On vrifie que l'im3d et l'image ont la mme taille */
  nbligIm3d = im3d->header.nblig;
  nbcolIm3d = im3d->header.nbcol;

  /* Slide the DTM */
  if (mnt->type == DTM_SLIDING) {
    dtm_slideDtm (mnt, RobotRed_2_Start);
  }
  
  /* Initialisations diverses et variées */
  date = mnt->nb_im3d + 1;
  nb_pts3d_out = 0;
  
  nblig_mnt = mnt->nblig;
  nbcol_mnt = mnt->nbcol;

  if (color != 0 && im3d->local_tab == NULL)
	  fprintf(stderr, "Warning : color computation requested, but local_tab is \
					   NULL!!\n");
  
  /* Boucle sur les points de l'im3d */
  pt3d = im3d->points_tab;
  luminance = im3d->local_tab;
  for (j = 0; j < nbcolIm3d; j++) {
    for (i = 0; i < nbligIm3d; i++) {
      
      if (pt3d->state == GOOD_PT3D) {
	i_mnt = (pt3d->coord_1 - mnt->x_origine) / mnt->scale_x;
	j_mnt = (pt3d->coord_2 - mnt->y_origine) / mnt->scale_y;
	
	if ((i_mnt < 0) || (i_mnt >= nblig_mnt)
	    || (j_mnt < 0) || (j_mnt >= nbcol_mnt))
	  nb_pts3d_out++;
	else {
	  cell = dtm_one_cell (mnt, i_mnt, j_mnt);
	  dtm_fusePoint (cell, pt3d, date, 0.0, color, fusionType,
			im3d->local_tab == NULL ? 0 :  (unsigned char) *luminance);
	  mnt->nb_cells++;
	}
      }
      pt3d++;
      luminance++;
    }
  }
  
/*   printf ("%d points out\n", nb_pts3d_out); */

  mnt->cells = CELLS_DETERMINEES;
  if (color != 0 && im3d->local_tab != NULL)
    mnt->luminance = LUMINANCES_DETERMINEES;
  mnt->nb_im3d++;
  
  return (DTM_OK);
}

