/* Les includes génériques de C */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

/* Les structures, protos et erreurs de la lib */
#include "libdtm.h"

/**********************************************************************
 *
 * Given a DTM, a camera model (intrinsinc calibration matrix and number of
 * pixels), and a 3D position of the camera in the DTM frame, returns the 3D
 * image perceived by the camera (maxDistance is the maximum perceived
 * distance). 
 *
 * Principle of the algorithm : 
 *
 *
 * world frame = DTM frame
 **********************************************************************/

DATA_IM3D*	
dtm_simulStereo (DTM*		dtm,
		 T3D*		cam2world,
		 double*	intrinsincParams,
		 int		nbLinesInit,
		 int		nbColumnsInit,
		 int		reduction,
		 double		maxDistance)

{
  int		nbLines;
  int		nbColumns;	/* The size of the image (after reduction) */
  int		i, j;		/* To loop on the pixels of the image */
  int		iCam, jCam;	/* Vertical projection of the camera in the
				   DTM */ 
  int		iEnd, jEnd;	/* Vertical projection of the end of one
				   camera ray in the DTM */ 

  DATA_IM3D*	im3d;		/* What is created, filled and returned by the 
				 function */
  DATA_PT3D*	point3d;	/* A point of the im3d */
  double	i0, j0;
  double	alpha_i, alpha_j; /* Camera calibration parameters */

  double	theta, phi;	/* Angles that define a line in the camera
				   frame */ 
  double	alphaX, alphaY, alphaZ;
				/* Coefficients of the parametric equation of
				   a line in the camera frame */
  double	betaX, betaY, betaZ;
				/* Coefficients of the parametric equation of
				   a line in the DTM frame */
  double	parameter;	/* To specify a point on the line */

  double	x1, y1, z1;	/* Coordinates of the intersection of the
				   camera ray and the DTM in the world frame */
  T3D*		world2cam;	/* To convert the corrdinates in the sensor frame */

  int		intersection;	/* Flag to indicate whether the line
				   intersects the DTM of not */
  int		iInter, jInter;	/* The coordinate of the intersection in the DTM */

  /* Creates the 3D image */
  nbLines = nbLinesInit / reduction;
  nbColumns = nbColumnsInit / reduction;
  if (cree_data_im3d (&im3d, nbLines, nbColumns, 0, 0, 0) != DTM_OK)
    return (NULL);
  /* Initializes various fields of the DATA_IM3D */
  im3d->header.capteur = CAPTEUR_STEREO_SIMULATION;
  im3d->header.repere = REPERE_CAPTEUR;
  im3d->header.coordonnees = COORDONNEES_CARTESIENNES;
  if (reduction != 1) 
    im3d->header.reduction = REDUCTION;
  else
    im3d->header.reduction = NIL_REDUCTION;
  im3d->header.sous_image = NIL_SOUS_IMAGE;
  im3d->header.nblig_initial = nbLinesInit;
  im3d->header.nbcol_initial = nbColumnsInit;
  im3d->header.nblig = nbLines;
  im3d->header.nbcol = nbColumns;
  im3d->header.i_min = 0;
  im3d->header.j_min = 0;
  im3d->header.i_max = nbLines;
  im3d->header.j_max = nbColumns;
  im3d->header.taux_reduction = reduction;
  /* Copy of the intrinsinc matrix */
  for (i = 0; i < 16; i++)
    im3d->header.mat_intrinseque[i] = *(intrinsincParams + i);

  /* Get the projection of the center of the camera in the DTM */
  iCam = (cam2world->euler.euler[3] - dtm->x_origine) / dtm->scale_x;
  jCam = (cam2world->euler.euler[4] - dtm->y_origine) / dtm->scale_y;

  /* Verify that the cameras is within the DTM, at a positive altitude */
  if ((iCam < 0) || (iCam > dtm->nblig) || (jCam < 0) || (jCam > dtm->nbcol)) {
    printf ("The camera is out of the DTM !\n");
    dtmError (ERR_DTM_BAD_PARAMS);
    return (NULL);
  }
  if ((dtm->cells_tab + (iCam * dtm->nbcol) + jCam)->current_info.z_moyen 
      > cam2world->euler.euler[5]) {
    printf ("The camera is below the ground surface !\n");
    dtmError (ERR_DTM_BAD_PARAMS);
    return (NULL);
  }

   t3dConvertTo(T3D_MATRIX, cam2world);

  /* Get the calibration params */
  i0 = *(intrinsincParams + 2) / reduction;
  j0 = *(intrinsincParams + 6) / reduction;
  alpha_i = *(intrinsincParams) / reduction;
  alpha_j = *(intrinsincParams + 5) / reduction;

  /* Compute world2cam */
  world2cam = t3dInvert (cam2world);
  t3dConvertTo(T3D_MATRIX, world2cam);

  /* Now loop on every pixel of the camera */
  point3d = im3d->points_tab; 
  for (i = 0; i < nbLines; i++) {
    for (j = 0; j < nbColumns; j++) {
      /* Each (i,j) couple defines a line by two angles */ 
      theta = atan ( (j0 - j) / alpha_j);
      phi = atan (cos(theta) * ( (i0 - i) / alpha_i));
      /* The coeficients of its parametric equation in the camera frame are
	 the following */  
      alphaX = cos(phi) * cos(theta);
      alphaY = cos(phi) * sin(theta);
      alphaZ = sin(phi);

      /* And in the DTM frame */
      betaX = alphaX * cam2world->matrix.matrix[0]
	      + alphaY * cam2world->matrix.matrix[1]
	      + alphaZ * cam2world->matrix.matrix[2];
      betaY = alphaX * cam2world->matrix.matrix[4]
	      + alphaY * cam2world->matrix.matrix[5]
	      + alphaZ * cam2world->matrix.matrix[6];
      betaZ = alphaX * cam2world->matrix.matrix[8]
	      + alphaY * cam2world->matrix.matrix[9]
	      + alphaZ * cam2world->matrix.matrix[10];
      betaZ = alphaX * cam2world->matrix.matrix[8]
	      + alphaY * cam2world->matrix.matrix[9]
	      + alphaZ * cam2world->matrix.matrix[10];

      /* Compute the coordinates of the line end in the world frame */
      x1 = maxDistance * betaX + cam2world->matrix.matrix[3];
      y1 = maxDistance * betaY + cam2world->matrix.matrix[7];
      z1 = maxDistance * betaZ + cam2world->matrix.matrix[11];

      /* Compute the line end projection coordinates in the dtm */
      iEnd = (x1 - dtm->x_origine) / dtm->scale_x;
      jEnd = (y1 - dtm->y_origine) / dtm->scale_y;

      /* Get the cell where the dtm and the line intersect */
      intersection = 0;
      dtm_intersectLineDtm (dtm, iCam, jCam, iEnd, jEnd, 
			    betaX, cam2world->matrix.matrix[3], betaZ, cam2world->matrix.matrix[11], 
			    &intersection, &iInter, &jInter);
      if (intersection != 0) {
	/* We have an intersection between the line and the DTM, at the
	   elevation of (iInter,jInter). The parameter value of the point is
	   retrieved with this elevation, and the coordinates (x,y) are
	   deduded from this parameter */
	z1 = (dtm->cells_tab + (iInter * dtm->nbcol) + jInter)->current_info.z_moyen;
	parameter = (z1 - cam2world->matrix.matrix[11]) / betaZ;
	x1 = parameter * betaX + cam2world->matrix.matrix[3];
	y1 = parameter * betaY + cam2world->matrix.matrix[7];

	/* Set the point coordinates in the sensor frame */
	point3d->state = GOOD_PT3D;
	point3d->coordonnees = COORDONNEES_CARTESIENNES;
	point3d->repere = REPERE_CAPTEUR;
	
	point3d->coord_1 = (x1 * world2cam->matrix.matrix[0]
			    + y1 * world2cam->matrix.matrix[1]
			    + z1 * world2cam->matrix.matrix[2])
			   + world2cam->matrix.matrix[3];
	point3d->coord_2 = (x1 * world2cam->matrix.matrix[4]
			    + y1 * world2cam->matrix.matrix[5]
			    + z1 * world2cam->matrix.matrix[6])
			   + world2cam->matrix.matrix[7];
	point3d->coord_3 = (x1 * world2cam->matrix.matrix[8]
			    + y1 * world2cam->matrix.matrix[9]
			    + z1 * world2cam->matrix.matrix[10])
			   + world2cam->matrix.matrix[11];
	
      }
      else
	point3d->state = BAD_PT3D;

      point3d++;
    }
  }
  
  /* Now we're done */
  im3d->header.points = POINTS_CALCULES;
  return (im3d);
}


/**********************************************************************
 *
 * Bressenham
 *
 **********************************************************************/

void 
dtm_intersectLineDtm (DTM*	dtm, 
		      int	i1,
		      int	j1,
		      int	i2,
		      int	j2, 
		      double	betaX,
		      double	translX,
		      double	betaZ,
		      double	translZ, 
		      int*	intersection, 
		      int*	iInter,
		      int*	jInter)

{
  int	xdelta;			/* The change in x coordinates */
  int	ydelta;			/* The change in y coordinates */
  int	xstep;			/* The change to make in the x coordinate in
				 * each step */
  int	ystep;			/* The change to make in the y coordinate in
				 * each step */
  int	change;			/* The amount that the x or y coordinate has
				 * changed */

  double	k;		/* The line parameter value for an intersection */
  double	x, z;

  DTM_CELL*	cell;
  int		nblig;
  int		nbcol;

  nblig = dtm->nblig;
  nbcol = dtm->nbcol;

  xdelta = i2 - i1;		/* Calculate the change in x coordinates */
  ydelta = j2 - j1;		/* Calculate the change in y coordinates */
  if (xdelta < 0) {		/* The line will be drawn from right to left */
    xdelta = -xdelta;
    xstep = -1;
  } else			/* The line will be drawn from left to right */
    xstep = 1;
  if (ydelta < 0) {		/* The line will be drawn from bottom to top */
    ydelta = -ydelta;
    ystep = -1;
  } else			/* The line will be drawn from top to bottom */
    ystep = 1;
  if (xdelta > ydelta) {	/* x changes quicker than y */
    change = xdelta >> 1;	/* change set to twice the value of xdelta */
    while (i1 != i2) {	/* Draw until the terminating dot is reached */
      
      /* (i1, j1) is a point of the line. Compute the corresponding line
	 altitude */ 
      if ((i1 > 0) && (i1 < nblig) && (j1 > 0) && (j1 < nbcol)) {
	x = (i1 + 0.5) * dtm->scale_x + dtm->x_origine;
	k = (x - translX) / betaX;
	z = k * betaZ + translZ;
	cell = dtm->cells_tab + (i1 * dtm->nbcol) + j1;
	if (z <= cell->current_info.z_moyen) {
	  /* The line went through the DTM surface */
	  *intersection = 1;
	  *iInter = i1;
	  *jInter = j1;
	  return;
	}
      }
      else {
	/* The point is out of the dtm */
	*intersection = 0;
	return;
      }
      
      i1 += xstep;		/* Update x coordinate */
      change += ydelta;		/* Update change */
      if (change > xdelta)
	/*
	 * If change is large enough to update the y coordinate
	 */
	{
	  j1 += ystep;		/* Update the y coordinate */
	  change -= xdelta;	/* Reset change */
	}
    }
  } else {			/* y changes quicker than x */
    change = ydelta >> 1;	/* change set to twice the value of ydelta */
    while (j1 != j2) {		/* Draw until the terminating dot is reached */

      /* (i1, j1) is a point of the line. Compute the corresponding line
	 altitude */ 
      if ((i1 > 0) && (i1 < nblig) && (j1 > 0) && (j1 < nbcol)) {
	x = (i1 + 0.5) * dtm->scale_x + dtm->x_origine;
	k = (x - translX) / betaX;
	z = k * betaZ + translZ;
	cell = dtm->cells_tab + (i1 * dtm->nbcol) + j1;
	if (z <= cell->current_info.z_moyen) {
	  /* The line went through the DTM surface */
	  *intersection = 1;
	  *iInter = i1;
	  *jInter = j1;
	  return;
	}
      }
      else {
	/* The point is out of the dtm */
	*intersection = 0;
	return;
      }

      j1 += ystep;		/* Update y coordinate */
      change += xdelta;		/* Update change */
      if (change > ydelta)
	/* If change is large enough to update the x coordinate */
	{
	  i1 += xstep;		/* Update the x coordinate */
	  change -= ydelta;	/* Reset change */
	}
    }
  }

  return;
}
